# pyflyby/_dbg.py.
# Copyright (C) 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2018 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT


import builtins
from   contextlib               import contextmanager
import errno
from   functools                import wraps
import os
import pwd
import signal
import sys
import time
from   types                    import CodeType, FrameType, TracebackType

from   collections.abc          import Callable

from   pyflyby._file            import Filename


"""
Used by wait_for_debugger_to_attach to record whether we're waiting to attach,
and if so what.
"""
_waiting_for_debugger = None


_ORIG_SYS_EXCEPTHOOK = sys.excepthook

def _reset_excepthook():
    if _ORIG_SYS_EXCEPTHOOK:
        sys.excepthook = _ORIG_SYS_EXCEPTHOOK
        return True
    return False


def _override_excepthook(hook):
    """
    Override sys.excepthook with `hook` but also support resetting.

    Users should call this function instead of directly overiding
    sys.excepthook. This is helpful in resetting sys.excepthook in certain cases.
    """
    global _ORIG_SYS_EXCEPTHOOK
    _ORIG_SYS_EXCEPTHOOK = hook
    sys.excepthook = hook


class _NoTtyError(Exception):
    pass


_memoized_dev_tty_fd = Ellipsis
def _dev_tty_fd():
    """
    Return a file descriptor opened to /dev/tty.
    Memoized.
    """
    global _memoized_dev_tty_fd
    if _memoized_dev_tty_fd is Ellipsis:
        try:
            _memoized_dev_tty_fd = os.open("/dev/tty", os.O_RDWR)
        except OSError:
            _memoized_dev_tty_fd = None
    if _memoized_dev_tty_fd is None:
        raise _NoTtyError
    return _memoized_dev_tty_fd


def tty_is_usable():
    """
    Return whether /dev/tty is usable.

    In interactive sessions, /dev/tty is usable; in non-interactive sessions,
    /dev/tty is not usable::

      $ ssh -t localhost py -q pyflyby._dbg.tty_is_usable
      True

      $ ssh -T localhost py -q pyflyby._dbg.tty_is_usable
      False

    tty_is_usable() is useful for deciding whether we are in an interactive
    terminal.  In an interactive terminal we can enter the debugger directly;
    in a non-interactive terminal, we need to wait_for_debugger_to_attach.

    Note that this is different from doing e.g. isatty(0).  isatty would
    return False if a program was piped, even though /dev/tty is usable.
    """
    try:
        _dev_tty_fd()
        return True
    except _NoTtyError:
        return False


@contextmanager
def _FdCtx(target_fd, src_fd):
    assert target_fd != src_fd
    saved_fd = os.dup(target_fd)
    assert saved_fd > 2, "saved_fd == %d" % (saved_fd,)
    assert saved_fd != target_fd and saved_fd != src_fd
    os.dup2(src_fd, target_fd)
    try:
        yield
    finally:
        os.dup2(saved_fd, target_fd)


_in_StdioCtx = []

@contextmanager
def _StdioCtx(tty="/dev/tty"):
    '''
    Within the context, force fd {0, 1, 2}, sys.__{stdin,stdout,stderr}__,
    sys.{stdin,stdout,stderr} to fd.  This allows us to use the debugger even
    if stdio is otherwise redirected.

    :type tty:
      ``int`` or ``str``
    :param tty:
      Tty to use.  Either a file descriptor or a name of a tty.
    '''
    from ._interactive import UpdateIPythonStdioCtx
    to_close = None
    if isinstance(tty, int):
        fd = tty
    elif isinstance(tty, str):
        if tty == "/dev/tty":
            fd = _dev_tty_fd()
        else:
            fd = os.open(tty, os.O_RDWR)
            to_close = fd
    else:
        raise TypeError("_StdioCtx(): tty should be an int or str")
    if _in_StdioCtx and _in_StdioCtx[-1] == fd:
        # Same context; do nothing.
        assert to_close is None
        yield
        return
    if not fd > 2:
        raise ValueError("_StdioCtx: unsafe to use fd<=2; fd==%d" % (fd,))
    _in_StdioCtx.append(fd)
    saved_stdin    = sys.stdin
    saved_stdin__  = sys.__stdin__
    saved_stdout   = sys.stdout
    saved_stdout__ = sys.__stdout__
    saved_stderr   = sys.stderr
    saved_stderr__ = sys.__stderr__
    try:
        sys.stdout.flush(); sys.__stdout__.flush()
        sys.stderr.flush(); sys.__stderr__.flush()
        from ._util import nested
        with nested(_FdCtx(0, fd), _FdCtx(1, fd), _FdCtx(2, fd)):
            with nested(os.fdopen(0, 'r'),
                        os.fdopen(1, 'w'),
                        os.fdopen(2, 'w', 1)) as (fd0, fd1, fd2):
                sys.stdin  = sys.__stdin__  = fd0
                sys.stdout = sys.__stdout__ = fd1
                sys.stderr = sys.__stderr__ = fd2
                # Update IPython's stdin/stdout/stderr temporarily.
                with UpdateIPythonStdioCtx():
                    yield
    finally:
        assert _in_StdioCtx and _in_StdioCtx[-1] == fd
        _in_StdioCtx.pop(-1)
        sys.stdin      = saved_stdin
        sys.__stdin__  = saved_stdin__
        sys.stdout     = saved_stdout
        sys.__stdout__ = saved_stdout__
        sys.stderr     = saved_stderr
        sys.__stderr__ = saved_stderr__
        if to_close is not None:
            try:
                os.close(to_close)
            except (OSError, IOError):
                pass


@contextmanager
def _ExceptHookCtx():
    '''
    Context manager that restores ``sys.excepthook`` upon exit.
    '''
    saved_excepthook = sys.excepthook
    try:
        # TODO: should we set sys.excepthook = sys.__excepthook__ ?
        yield
    finally:
        sys.excepthook = saved_excepthook


@contextmanager
def _DisplayHookCtx():
    '''
    Context manager that resets ``sys.displayhook`` to the default value upon
    entry, and restores the pre-context value upon exit.
    '''
    saved_displayhook = sys.displayhook
    try:
        sys.displayhook = sys.__displayhook__
        yield
    finally:
        sys.displayhook = saved_displayhook


def print_traceback(*exc_info):
    """
    Print a traceback, using IPython's ultraTB if possible.

    Output goes to /dev/tty.

    :param exc_info:
      3 arguments as returned by sys.exc_info().
    """
    from pyflyby._interactive import print_verbose_tb
    if not exc_info:
        exc_info = sys.exc_info()
    with _StdioCtx():
        print_verbose_tb(*exc_info)


@contextmanager
def _DebuggerCtx(tty="/dev/tty"):
    """
    A context manager that sets up the environment (stdio, sys hooks) for a
    debugger, initializes IPython if necessary, and creates a debugger instance.

    :return:
      Context manager that yields a Pdb instance.
    """
    from pyflyby._interactive import new_IPdb_instance
    with _StdioCtx(tty):
        with _ExceptHookCtx():
            with _DisplayHookCtx():
                pdb = new_IPdb_instance()
                pdb.reset()
                yield pdb


def _get_caller_frame():
    '''
    Get the closest frame from outside this module.

    :rtype:
      ``FrameType``
    '''
    this_filename = _get_caller_frame.__code__.co_filename
    f = sys._getframe()
    while (f.f_back and (
            f.f_code.co_filename == this_filename or
            (f.f_back.f_back and
             f.f_code.co_filename == "<string>" and
             f.f_code.co_name == "<module>" and
             f.f_back.f_code.co_filename == this_filename))):
        f = f.f_back
    if f.f_code.co_filename == "<string>" and f.f_code.co_name == "<module>":
        # Skip an extra string eval frame for attaching a debugger.
        # TODO: pass in a frame or maximum number of string frames to skip.
        # We shouldn't skip "<string>" if it comes from regular user code.
        f = f.f_back
    return f


def _prompt_continue_waiting_for_debugger():
    """
    Prompt while exiting the debugger to get user opinion on keeping the
    process waiting for debugger to attach.
    """
    count_invalid = 0
    max_invalid_entries = 3
    while count_invalid < max_invalid_entries:
        sys.stdout.flush()
        response = input("Keep the process running for debugger to be "
                            "attached later ? (y)es/(n)o\n")

        if response.lower() in ('n', 'no'):
            global _waiting_for_debugger
            _waiting_for_debugger = None
            break

        if response.lower() not in ('y', 'yes'):
            print("Invalid response: {}".format(repr(response)))
            count_invalid += 1
        else:
            break
    else:
        print("Exiting after {} invalid responses.".format(max_invalid_entries))
        _waiting_for_debugger = None
        # Sleep for a fraction of second for the print statements to get printed.
        time.sleep(0.01)


def _debug_exception(*exc_info, **kwargs):
    """
    Debug an exception -- print a stack trace and enter the debugger.

    Suitable to be assigned to sys.excepthook.
    """
    from pyflyby._interactive import print_verbose_tb
    tty = kwargs.pop("tty", "/dev/tty")
    debugger_attached = kwargs.pop("debugger_attached", False)
    if kwargs:
        raise TypeError("debug_exception(): unexpected kwargs %s"
                        % (', '.join(sorted(kwargs.keys()))))
    if not exc_info:
        exc_info = sys.exc_info()
    if len(exc_info) == 1 and type(exc_info[0]) is tuple:
        exc_info = exc_info[0]
    if len(exc_info) == 1 and type(exc_info[0]) is TracebackType:
        # Allow the input to be just the traceback.  The exception instance is
        # only used for printing the traceback.  It's not needed by the
        # debugger.
        # We don't know the exception in this case.  For now put "", "".  This
        # will cause print_verbose_tb to include a line with just a colon.
        # TODO: avoid that line.
        exc_info = ("", "", exc_info)
    if exc_info[1]:
        # Explicitly set sys.last_value / sys.last_exc to ensure they are available
        # in the debugger. One use case is that this allows users to call 
        # pyflyby.saveframe() within the debugger.
        if sys.version_info < (3, 12):
            sys.last_value = exc_info[1]
        else:
            sys.last_exc = exc_info[1]

    with _DebuggerCtx(tty=tty) as pdb:
        if debugger_attached:
            # If debugger is attached to the process made waiting by
            # 'wait_for_debugger_to_attach', check with the user whether to
            # keep the process waiting for debugger to attach.
            pdb.postloop = _prompt_continue_waiting_for_debugger
        print_verbose_tb(*exc_info)
        # Starting Py3.13, pdb.interaction() supports chained exceptions in case
        # exception (and not traceback) is specified. This support is backported
        # to IPython8.16 for earlier Python versions. So the conditions where
        # chained exceptions won't be supported from here would be with the
        # Python version < 3.13 and ipython not installed, or IPython's version
        # is lesser than 8.16.
        tb_or_exc = exc_info[2]
        if sys.version_info < (3, 13):
            # Check if the instance is of IPython's Pdb and its version.
            try:
                import IPython
                if IPython.version_info >= (8, 16):
                    from IPython.core.debugger import Pdb as IPdb
                    # This is expected to be True, hence just a safe check.
                    if isinstance(pdb, IPdb):
                        tb_or_exc = exc_info[1]
            except ModuleNotFoundError:
                pass
        else:
            tb_or_exc = exc_info[1]
        pdb.interaction(None, tb_or_exc)


def _debug_code(arg, globals=None, locals=None, auto_import=True, tty="/dev/tty"):
    """
    Run code under the debugger.

    :type arg:
      ``str``, ``Callable``, ``CodeType``, ``PythonStatement``, ``PythonBlock``,
      ``FileText``
    """
    if globals is None or locals is None:
        caller_frame = _get_caller_frame()
        if globals is None:
            globals = caller_frame.f_globals
        if locals is None:
            locals = caller_frame.f_locals
        del caller_frame
    with _DebuggerCtx(tty=tty) as pdb:
        print("Entering debugger.  Use 'n' to step, 'c' to run, 'q' to stop.")
        print("")
        from ._parse import PythonStatement, PythonBlock, FileText

        if isinstance(arg, (str, PythonStatement, PythonBlock, FileText)):
            # Compile the block so that we can get the right compile mode.
            arg = PythonBlock(arg)
            # TODO: enter text into linecache
            autoimp_arg = arg
            code = arg.compile()
        elif isinstance(arg, CodeType):
            autoimp_arg = arg
            code = arg
        elif isinstance(arg, Callable):
            # TODO: check argspec to make sure it's a zero-arg callable.
            code = arg.__code__
            autoimp_arg = code
        else:
            raise TypeError(
                "debug_code(): expected a string/callable/lambda; got a %s"
                % (type(arg).__name__,))
        if auto_import:
            from ._autoimp import auto_import as auto_import_f
            auto_import_f(autoimp_arg, [globals, locals])
        return pdb.runeval(code, globals=globals, locals=locals)


_CURRENT_FRAME = object()


def debugger(*args, **kwargs):
    '''
    Entry point for debugging.

    ``debugger()`` can be used in the following ways::

    1. Breakpoint mode, entering debugger in executing code::
         >> def foo():
         ..     bar()
         ..     debugger()
         ..     baz()

       This allow stepping through code after the debugger() call - i.e. between
       bar() and baz().  This is similar to 'import pdb; pdb.set_trace()'::

    2. Debug a python statement::

         >> def foo(x):
         ..     ...
         >> X = 5

         >> debugger("foo(X)")

       The auto-importer is run on the given python statement.

    3. Debug a callable::

         >> def foo(x=5):
         ..     ...

         >> debugger(foo)
         >> debugger(lambda: foo(6))

    4. Debug an exception::

         >> try:
         ..     ...
         .. except:
         ..     debugger(sys.exc_info())


    If the process is waiting on for a debugger to attach to debug a frame or
    exception traceback, then calling debugger(None) will debug that target.
    If it is frame, then the user can step through code.  If it is an
    exception traceback, then the debugger will operate in post-mortem mode
    with no stepping allowed.  The process will continue running after this
    debug session's "continue".

    ``debugger()`` is suitable to be called interactively, from scripts, in
    sys.excepthook, and in signal handlers.

    :param args:
      What to debug:
        - If a string or callable, then run it under the debugger.
        - If a frame, then debug the frame.
        - If a traceback, then debug the traceback.
        - If a 3-tuple as returned by sys.exc_info(), then debug the traceback.
        - If the process is waiting to for a debugger to attach, then attach
          the debugger there.  This is only relevant when an external process
          is attaching a debugger.
        - If nothing specified, then enter the debugger at the statement
          following the call to debug().
    :kwarg tty:
      Tty to connect to.  If ``None`` (default): if /dev/tty is usable, then
      use it; else call wait_for_debugger_to_attach() instead (unless
      wait_for_attach==False).
    :kwarg on_continue:
      Function to call upon exiting the debugger and continuing with regular
      execution.
    :kwarg wait_for_attach:
      Whether to wait for a remote terminal to attach (with 'py -d PID').
      If ``True``, then always wait for a debugger to attach.
      If ``False``, then never wait for a debugger to attach; debug in the
      current terminal.
      If unset, then defaults to true only when ``tty`` is unspecified and
      /dev/tty is not usable.
    :kwarg background:
      If ``False``, then pause execution to debug.
      If ``True``, then fork a process and wait for a debugger to attach in the
      forked child.
    '''
    from ._parse import PythonStatement, PythonBlock, FileText
    if len(args) in {1, 2}:
        arg = args[0]
    elif len(args) == 0:
        arg = None
    else:
        arg = args
    tty             = kwargs.pop("tty"            , None)
    on_continue     = kwargs.pop("on_continue"    , lambda: None)
    globals         = kwargs.pop("globals"        , None)
    locals          = kwargs.pop("locals"         , None)
    wait_for_attach = kwargs.pop("wait_for_attach", Ellipsis)
    background      = kwargs.pop("background"     , False)
    _debugger_attached = False
    if kwargs:
        raise TypeError("debugger(): unexpected kwargs %s"
                        % (', '.join(sorted(kwargs))))
    if arg is None and tty is not None and wait_for_attach != True:
        # If _waiting_for_debugger is not None, then attach to that
        # (whether it's a frame, traceback, etc).
        global _waiting_for_debugger
        arg = _waiting_for_debugger
        _debugger_attached = True
    if arg is None:
        # Debug current frame.
        arg = _CURRENT_FRAME
    if arg is _CURRENT_FRAME:
        arg = _get_caller_frame()
    if background:
        # Fork a process and wait for a debugger to attach in the background.
        # Todo: implement on_continue()
        wait_for_debugger_to_attach(arg, background=True)
        return
    if wait_for_attach == True:
        wait_for_debugger_to_attach(arg)
        return
    if tty is None:
        if tty_is_usable():
            tty = "/dev/tty"
        elif wait_for_attach != False:
            # If the tty isn't usable, then default to waiting for the
            # debugger to attach from another (interactive) terminal.
            # Todo: implement on_continue()
            # TODO: capture globals/locals when relevant.
            wait_for_debugger_to_attach(arg)
            return
    if isinstance(
        arg, (str, PythonStatement, PythonBlock, FileText, CodeType, Callable)
    ):
        _debug_code(arg, globals=globals, locals=locals, tty=tty)
        on_continue()
        return
    if (isinstance(arg, TracebackType) or
        type(arg) is tuple and len(arg) == 3 and type(arg[2]) is TracebackType):
        _debug_exception(arg, tty=tty, debugger_attached=_debugger_attached)
        on_continue()
        return
    import threading
    # If `arg` is an instance of `tuple` that contains
    # `threading.ExceptHookArgs`, extract the exc_info from it.
    if type(arg) is tuple and len(arg) == 1 and type(arg[0]) is threading.ExceptHookArgs:
        arg = arg[0][:3]
        _debug_exception(arg, tty=tty, debugger_attached=_debugger_attached)
        on_continue()
        return
    if not isinstance(arg, FrameType):
        raise TypeError(
            "debugger(): expected a frame/traceback/str/code; got %s"
            % (arg,))
    frame = arg
    if globals is not None or locals is not None:
        raise NotImplementedError(
            "debugger(): globals/locals only relevant when debugging code")
    pdb_context = _DebuggerCtx(tty)
    pdb = pdb_context.__enter__()
    print("Entering debugger.  Use 'n' to step, 'c' to continue running, 'q' to quit Python completely.")
    def set_continue():
        # Continue running code outside the debugger.
        pdb.stopframe = pdb.botframe
        pdb.returnframe = None
        sys.settrace(None)
        print("Continuing execution.")
        pdb_context.__exit__(None, None, None)
        on_continue()
    def set_quit():
        # Quit the program.  Note that if we're inside IPython, then this
        # won't actually exit IPython.  We do want to call the context
        # __exit__ here to make sure we restore sys.displayhook, etc.
        # TODO: raise something else here if in IPython
        pdb_context.__exit__(None, None, None)
        raise SystemExit("Quitting as requested while debugging.")
    pdb.set_continue = set_continue
    pdb.set_quit = set_quit
    pdb.do_EOF = pdb.do_continue
    pdb.set_trace(frame)
    # Note: set_trace() installs a tracer and returns; that means we can't use
    # context managers around set_trace(): the __exit__() would be called
    # right away, not after continuing/quitting.
    # We also want this to be the very last thing called in the function (and
    # not in a nested function).  This way the very next thing the user sees
    # is his own code.



_cached_py_commandline = None
def _find_py_commandline():
    global _cached_py_commandline
    if _cached_py_commandline is not None:
        return _cached_py_commandline
    import pyflyby
    pkg_path = Filename(pyflyby.__path__[0]).real
    assert pkg_path.base == "pyflyby"
    d = pkg_path.dir
    if d.base == "bin":
        # Running from source tree
        bindir = d
    else:
        # Installed by setup.py
        while d.dir != d:
            d = d.dir
            bindir = d / "bin"
            if bindir.exists:
                break
        else:
            raise ValueError(
                "Couldn't find 'py' script: "
                "couldn't find 'bin' dir from package path %s" % (pkg_path,))
    candidate = bindir / "py"
    if not candidate.exists:
        raise ValueError(
            "Couldn't find 'py' script: expected it at %s" % (candidate,))
    if not candidate.isexecutable:
        raise ValueError(
            "Found 'py' script at %s but it's not executable" % (candidate,))
    _cached_py_commandline = candidate
    return candidate



class DebuggerAttachTimeoutError(Exception):
    pass


def _sleep_until_debugger_attaches(arg, timeout=86400):
    assert arg is not None
    global _waiting_for_debugger
    try:
        deadline = time.time() + timeout
        _waiting_for_debugger = arg
        while _waiting_for_debugger is not None:
            if time.time() > deadline:
                raise DebuggerAttachTimeoutError
            time.sleep(0.5)
    finally:
        _waiting_for_debugger = None


def wait_for_debugger_to_attach(arg, mailto=None, background=False, timeout=86400):
    """
    Send email to user and wait for debugger to attach.

    :param arg:
      What to debug.  Should be a sys.exc_info() result or a sys._getframe()
      result.
    :param mailto:
      Recipient to email.  Defaults to $USER or current user.
    :param background:
      If True, fork a child process.  The parent process continues immediately
      without waiting.  The child process waits for a debugger to attach, and
      exits when the debugging session completes.
    :param timeout:
      Maximum number of seconds to wait for user to attach debugger.
    """
    import traceback
    if background:
        originalpid = os.getpid()
        if os.fork() != 0:
            return
    else:
        originalpid = None
    try:
        # Reset the exception hook after the first exception.
        #
        # In case the code injected by the remote client causes some error in
        # the debugged process, another email is sent for the new exception. This can
        # lead to an infinite loop of sending mail for each successive exceptions
        # everytime a remote client tries to connect. Our process might never get
        # a chance to exit and the remote client might just hang.
        #
        if not _reset_excepthook():
            raise ValueError("Couldn't reset sys.excepthook. Aborting remote "
                             "debugging.")
        # Send email.
        _send_email_with_attach_instructions(arg, mailto, originalpid=originalpid)
        # Sleep until the debugger to attaches.
        _sleep_until_debugger_attaches(arg, timeout=timeout)
    except:
        traceback.print_exception(*sys.exc_info())
    finally:
        if background:
            # Exit.  Note that the original process already continued.
            # We do this in a 'finally' to make sure that we always exit
            # here.  We don't want to do cleanup actions (finally clauses,
            # atexit functions) in the parent, since that can affect the
            # parent (e.g. deleting temp files while the parent process is
            # still using them).
            os._exit(1)


def debug_on_exception(function, background=False):
    """
    Decorator that wraps a function so that we enter a debugger upon exception.
    """
    @wraps(function)
    def wrapped_function(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except:
            debugger(sys.exc_info(), background=background)
            raise
    return wrapped_function


def _send_email_with_attach_instructions(arg, mailto, originalpid):
    from   email.mime.text import MIMEText
    import smtplib
    import socket
    import traceback
    # Prepare variables we'll use in the email.
    d = dict()
    user = pwd.getpwuid(os.geteuid()).pw_name
    argv = ' '.join(sys.argv)
    d.update(
        argv       =argv                  ,
        argv_abbrev=argv[:40]             ,
        event      ="breakpoint"          ,
        exc        =None                  ,
        exctype    =None                  ,
        hostname   =socket.getfqdn()      ,
        originalpid=originalpid           ,
        pid        =os.getpid()           ,
        py         =_find_py_commandline(),
        time       =time.strftime("%Y-%m-%d %H:%M:%S %Z", time.localtime()),
        traceback  =None                  ,
        username   =user                  ,
    )
    tb = None
    frame = None
    stacktrace = None
    if isinstance(arg, FrameType):
        frame       = arg
        stacktrace = ''.join(traceback.format_stack(frame))
    elif isinstance(arg, TracebackType):
        frame = d['tb'].tb_frame
        stacktrace = ''.join(traceback.format_tb(arg))
    elif isinstance(arg, tuple) and len(arg) == 3 and isinstance(arg[2], TracebackType):
        d.update(
            exctype=arg[0].__name__,
            exc    =arg[1]         ,
            event  =arg[0].__name__,
        )
        tb = arg[2]
        while tb.tb_next:
            tb = tb.tb_next
        frame = tb.tb_frame
        stacktrace = ''.join(traceback.format_tb(arg[2])) + (
            "  %s: %s\n" % (arg[0].__name__, arg[1]))
    if not frame:
        frame = _get_caller_frame()
    d.update(
        function = frame.f_code.co_name    ,
        filename = frame.f_code.co_filename,
        line     = frame.f_lineno          ,
    )
    d.update(
        filename_abbrev = _abbrev_filename(d['filename']),
    )
    if tb:
        d['stacktrace'] = tb and ''.join("    %s\n" % (line,) for line in stacktrace.splitlines())
    # Construct a template for the email body.
    template = []
    template += [
        "While running {argv_abbrev}, {event} in {function} at {filename}:{line}",
        "",
        "Please run:",
        "    ssh -t {hostname} {py} -d {pid}",
        "",
    ]
    if d['originalpid']:
        template += [
            "As process {originalpid}, I have forked to process {pid} and am waiting for a debugger to attach."
        ]
    else:
        template += [
            "As process {pid}, I am waiting for a debugger to attach."
        ]
    template += [
        "",
        "Details:",
        "  Time             : {time}",
        "  Host             : {hostname}",
    ]
    if d['originalpid']:
        template += [
            "  Original process : {originalpid}",
            "  Forked process   : {pid}",
        ]
    else:
        template += [
            "  Process          : {pid}",
        ]
    template += [
        "  Username         : {username}",
        "  Command line     : {argv}",
    ]
    if d['exc']:
        template += [
            "  Exception        : {exctype}: {exc}",
        ]
    if d.get('stacktrace'):
        template += [
            "  Traceback        :",
            "{stacktrace}",
        ]
    # Build email body.
    email_body = '\n'.join(template).format(**d)
    # Print to stderr.
    prefixed = "".join("[PYFLYBY] %s\n" % line
                       for line in email_body.splitlines())
    sys.stderr.write(prefixed)
    # Send email.
    if mailto is None:
        mailto = os.getenv("USER") or user
    msg = MIMEText(email_body)
    msg['Subject'] = (
        "ssh {hostname} py -d {pid}"
        " # {event} in {argv_abbrev} in {function} at {filename_abbrev}:{line}"
        ).format(**d)
    msg['From'] = user
    msg['To'] = mailto
    s = smtplib.SMTP("localhost")
    s.sendmail(user, [mailto], msg.as_string())
    s.quit()


def _abbrev_filename(filename):
    splt = filename.rsplit("/", 4)
    if len(splt) >= 4:
        splt[:2] = ["..."]
    return '/'.join(splt)


def syscall_marker(msg):
    """
    Execute a dummy syscall that is visible in truss/strace.
    """
    try:
        s = ("/###        %s" % (msg,)).ljust(70)
        os.stat(s)
    except OSError:
        pass


_ORIG_PID = os.getpid()

def _signal_handler_debugger(signal_number, interrupted_frame):
    if os.getpid() != _ORIG_PID:
        # We're in a forked subprocess.  Ignore this SIGQUIT.
        return
    fd_tty = _dev_tty_fd()
    os.write(fd_tty, b"\nIntercepted SIGQUIT; entering debugger.  Resend ^\\ to dump core (and 'stty sane' to reset terminal settings).\n\n")
    frame = _get_caller_frame()
    enable_signal_handler_debugger(False)
    debugger(
        frame,
        on_continue=enable_signal_handler_debugger)
    signal.signal(signal.SIGQUIT, _signal_handler_debugger)


def enable_signal_handler_debugger(enable=True):
    r'''
    Install a signal handler for SIGQUIT so that Control-\ or external SIGQUIT
    enters debugger.  Suitable to be called from site.py.
    '''
    # Idea from bzrlib.breakin
    # (http://bazaar.launchpad.net/~bzr/bzr/trunk/annotate/head:/bzrlib/breakin.py)
    if enable:
        signal.signal(signal.SIGQUIT, _signal_handler_debugger)
    else:
        signal.signal(signal.SIGQUIT, signal.SIG_DFL)


def enable_exception_handler_debugger():
    '''
    Enable ``sys.excepthook = debugger`` so that we automatically enter
    the debugger upon uncaught exceptions.
    '''
    _override_excepthook(debugger)


# Handle SIGTERM with traceback+exit.
def _sigterm_handler(signum, frame):
    # faulthandler.dump_traceback(all_threads=True)
    import traceback
    traceback.print_stack()
    # raise SigTermReceived
    signal.signal(signum, signal.SIG_DFL)
    os.kill(os.getpid(), signum)
    os._exit(99) # shouldn't get here


def enable_sigterm_handler(on_existing_handler='raise'):
    """
    Install a handler for SIGTERM that causes Python to print a stack trace
    before exiting.

    :param on_existing_handler:
      What to do when a SIGTERM handler was already registered.
        - If ``"raise"``, then keep the existing handler and raise an exception.
        - If ``"keep_existing"``, then silently keep the existing handler.
        - If ``"warn_and_override"``, then override the existing handler and log a warning.
        - If ``"silently_override"``, then silently override the existing handler.
    """
    old_handler = signal.signal(signal.SIGTERM, _sigterm_handler)
    if old_handler == signal.SIG_DFL or old_handler == _sigterm_handler:
        return
    if on_existing_handler == "silently_override":
        return
    if on_existing_handler == "warn_and_override":
        from ._log import logger
        logger.warning("enable_sigterm_handler(): Overriding existing SIGTERM handler")
        return
    signal.signal(signal.SIGTERM, old_handler)
    if on_existing_handler == "keep_existing":
        return
    elif on_existing_handler == "raise":
        raise ValueError(
            "enable_sigterm_handler(on_existing_handler='raise'): SIGTERM handler already exists" + repr(old_handler))
    else:
        raise ValueError(
            "enable_sigterm_handler(): SIGTERM handler already exists, "
            "and invalid on_existing_handler=%r"
            % (on_existing_handler,))


def enable_faulthandler():
    try:
        import faulthandler
    except ImportError:
        pass
    else:
        # Print Python user-level stack trace upon SIGSEGV/etc.
        faulthandler.enable()


def add_debug_functions_to_builtins(*, add_deprecated: bool):
    """
    Install debugger(), etc. in the builtin global namespace.
    """
    functions_to_add = [
        'debugger',
        'debug_on_exception',
        'print_traceback',
    ]
    if add_deprecated:
        # DEPRECATED: In the future, the following will not be added to builtins.
        # Use debugger() instead.
        functions_to_add += [
            "breakpoint",
            "debug_exception",
            "debug_statement",
            "waitpoint",
        ]
    for name in functions_to_add:
        setattr(builtins, name, globals()[name])

# TODO: allow attaching remotely (winpdb/rpdb2) upon sigquit.  Or rpc like http://code.activestate.com/recipes/576515/
# TODO: http://sourceware.org/gdb/wiki/PythonGdb


def get_executable(pid):
    """
    Get the full path for the target process.

    :type pid:
      ``int``
    :rtype:
      `Filename`
    """
    uname = os.uname()[0]
    if uname == 'Linux':
        result = os.readlink('/proc/%d/exe' % (pid,))
    elif uname == 'SunOS':
        result = os.readlink('/proc/%d/path/a.out' % (pid,))
    else:
        # Use psutil to try to answer this.  This should also work for the
        # above cases too, but it's simple enough to implement it directly and
        # avoid this dependency on those platforms.
        import psutil
        result = psutil.Process(pid).exe()
    result = Filename(result).real
    if not result.isfile:
        raise ValueError("Couldn't get executable for pid %s" % (pid,))
    if not result.isreadable:
        raise ValueError("Executable %s for pid %s is not readable"
                         % (result, pid))
    return result


_gdb_safe_chars = (
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    r"0123456789,./-_=+:;'[]{}\|`~!@#%^&*()<>? ")

def _escape_for_gdb(string):
    """
    Escape a string to make it safe for passing to gdb.
    """
    result = []
    for char in string:
        if char in _gdb_safe_chars:
            result.append(char)
        else:
            result.append(r"\0{0:o}".format(ord(char)))
    return ''.join(result)


_memoized_dev_null = None
def _dev_null():
    """
    Return a file object opened for reading/writing to /dev/null.
    Memoized.

    :rtype:
      ``file``
    """
    global _memoized_dev_null
    if _memoized_dev_null is None:
        _memoized_dev_null = open("/dev/null", 'w+')
    return _memoized_dev_null


def inject(pid, statements, wait=True, show_gdb_output=False):
    """
    Execute ``statements`` in a running Python process.

    :type pid:
      ``int``
    :param pid:
      Id of target process
    :type statements:
      Iterable of strings
    :param statements:
      Python statements to execute.
    :return:
      Then process ID of the gdb process if ``wait`` is False; ``None`` if
      ``wait`` is True.
    """
    import subprocess
    os.kill(pid, 0) # raises OSError "No such process" unless pid exists
    if isinstance(statements, str):
        statements = (statements,)
    else:
        statements = tuple(statements)
    for statement in statements:
        if not isinstance(statement, str):
            raise TypeError(
                "Expected iterable of strings, not %r" % (type(statement),))
    # Based on
    # https://github.com/lmacken/pyrasite/blob/master/pyrasite/inject.py
    # TODO: add error checking
    # TODO: consider using lldb, especially on Darwin.
    gdb_commands = (
        [ 'PyGILState_Ensure()' ]
        + [ 'PyRun_SimpleString("%s")' % (_escape_for_gdb(statement),)
            for statement in statements ]
        + [ 'PyGILState_Release($1)' ])
    python_path = get_executable(pid)
    if "python" not in python_path.base:
        raise ValueError(
            "pid %s uses executable %s, which does not appear to be python"
            % (pid, python_path))
    # TODO: check that gdb is found and that the version is new enough (7.x)
    #
    # A note about --interpreter=mi: mi stands for Machine Interface and it's
    # the blessed way to control gdb from a pipe, since the output is much
    # easier to parse than the normal human-oriented output (it is also worth
    # noting that at the moment we are never parsig the output, but it's still
    # a good practice to use --interpreter=mi).
    command = (
        ['gdb', str(python_path), '-p', str(pid), '-batch', '--interpreter=mi']
        + [ '-eval-command=call %s' % (c,) for c in gdb_commands ])
    output = None if show_gdb_output else _dev_null()
    process = subprocess.Popen(command,
                               stdin=_dev_null(),
                               stdout=output,
                               stderr=output)
    if wait:
        retcode = process.wait()
        if retcode:
            raise Exception(
                "Gdb command %r failed (exit code %r)"
                % (command, retcode))
    else:
        return process.pid

import tty


# Copy of tty.setraw that does not set ISIG,
# in order to keep CTRL-C sending Keybord Interrupt.
def setraw_but_sigint(fd, when=tty.TCSAFLUSH):
    """Put terminal into a raw mode."""
    mode = tty.tcgetattr(fd)
    mode[tty.IFLAG] = mode[tty.IFLAG] & ~(
        tty.BRKINT | tty.ICRNL | tty.INPCK | tty.ISTRIP | tty.IXON
    )
    mode[tty.OFLAG] = mode[tty.OFLAG] & ~(tty.OPOST)
    mode[tty.CFLAG] = mode[tty.CFLAG] & ~(tty.CSIZE | tty.PARENB)
    mode[tty.CFLAG] = mode[tty.CFLAG] | tty.CS8
    mode[tty.LFLAG] = mode[tty.LFLAG] & ~(
        tty.ECHO | tty.ICANON | tty.IEXTEN
    )  # NOT ISIG HERE.
    mode[tty.CC][tty.VMIN] = 1
    mode[tty.CC][tty.VTIME] = 0
    tty.tcsetattr(fd, when, mode)


class Pty(object):
    def __init__(self):
        import pty
        self.master_fd, self.slave_fd = pty.openpty()
        self.ttyname = os.ttyname(self.slave_fd)

    def communicate(self):
        import tty
        import pty
        try:
            mode = tty.tcgetattr(pty.STDIN_FILENO)
            setraw_but_sigint(pty.STDIN_FILENO)
            restore = True
        except tty.error:
            restore = False
        try:
            pty._copy(self.master_fd)
        except KeyboardInterrupt:
            print('^C\r') # we need the \r because we are still in raw mode
        finally:
            if restore:
                tty.tcsetattr(pty.STDIN_FILENO, tty.TCSAFLUSH, mode)
        os.close(self.master_fd)


def process_exists(pid):
    """
    Return whether ``pid`` exists.

    :type pid:
      ``int``
    :rtype:
      ``bool``
    """
    try:
        os.kill(pid, 0)
        return True
    except OSError as e:
        if e.errno == errno.ESRCH:
            return False
        raise


def kill_process(pid, kill_signals):
    """
    Kill process ``pid`` using various signals.

    :param kill_signals:
      Sequence of (signal, delay) tuples.  Each signal is tried in sequence,
      waiting up to ``delay`` seconds before trying the next signal.
    """
    for sig, delay in kill_signals:
        start_time = time.time()
        try:
            os.kill(pid, sig)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return True
            raise
        deadline = start_time + delay
        while time.time() < deadline:
            if not process_exists(pid):
                return True
            time.sleep(0.05)


def attach_debugger(pid):
    """
    Attach command-line debugger to a running process.

    :param pid:
      Process id of target process.
    """
    import pyflyby
    import signal
    class SigUsr1(Exception):
        pass
    def sigusr1_handler(*args):
        raise SigUsr1
    signal.signal(signal.SIGUSR1, sigusr1_handler)
    terminal = Pty()
    pyflyby_lib_path = os.path.dirname(pyflyby.__path__[0])
    # Inject a call to 'debugger()' into target process.
    # Set on_continue to signal ourselves that we're done.
    on_continue = "lambda: __import__('os').kill(%d, %d)" % (os.getpid(), signal.SIGUSR1)

    # Use Python import machinery to import pyflyby from its directory.
    #
    # Adding the path to sys.path might have side effects. For e.g., a package
    # with the same name as a built-in module could exist in `pyflyby_dir`.
    # Adding `pyflyby_dir` to sys.path will make the package get imported from
    # `pyflyby_dir` instead of deferring this decision to the user Python
    # environment.
    #
    # As a concrete example, `typing` module is a package as well a built-in
    # module from Python version >= 3.5
    statements = [
        "loader = __import__('importlib').machinery.PathFinder.find_module("
        "fullname='pyflyby', path=['{pyflyby_dir}'])".format(
            pyflyby_dir=pyflyby_lib_path),
        "pyflyby = loader.load_module('pyflyby')"
    ]
    statements.append(
        ("pyflyby.debugger(tty=%r, on_continue=%s)"
         % (terminal.ttyname, on_continue))
        )

    gdb_pid = inject(pid, statements=";".join(statements), wait=False)
    # Fork a watchdog process to make sure we exit if the target process or
    # gdb process exits, and make sure the gdb process exits if we exit.
    parent_pid = os.getpid()
    watchdog_pid = os.fork()
    if watchdog_pid == 0:
        while True:
            try:
                if not process_exists(gdb_pid):
                    kill_process(
                        parent_pid,
                        [(signal.SIGUSR1, 5), (signal.SIGTERM, 15),
                         (signal.SIGKILL, 60)])
                    break
                if not process_exists(pid):
                    start_time = time.time()
                    os.kill(parent_pid, signal.SIGUSR1)
                    kill_process(
                        gdb_pid,
                        [(0, 5), (signal.SIGTERM, 15), (signal.SIGKILL, 60)])
                    kill_process(
                        parent_pid,
                        [(0, (5 + time.time() - start_time)),
                         (signal.SIGTERM, 15), (signal.SIGKILL, 60)])
                    break
                if not process_exists(parent_pid):
                    kill_process(
                        gdb_pid,
                        [(0, 5), (signal.SIGTERM, 15), (signal.SIGKILL, 60)])
                    break
                time.sleep(0.1)
            except KeyboardInterrupt:
                # if the user pressed CTRL-C the parent process is about to
                # die, so we will detect the death in the next iteration of
                # the loop and exit cleanly after killing also gdb
                pass
        os._exit(0)
    # Communicate with pseudo tty.
    try:
        terminal.communicate()
    except SigUsr1:
        print("\nDebugging complete.")
        pass


def remote_print_stack(pid, output=1):
    """
    Tell a target process to print a stack trace.

    This currently only handles the main thread.
    TODO: handle multiple threads.

    :param pid:
      PID of target process.
    :type output:
      ``int``, ``file``, or ``str``
    :param output:
      Output file descriptor.
    """
    # Interpret ``output`` argument as a file-like object, file descriptor, or
    # filename.
    if hasattr(output, 'write'): # file-like object
        output_fh = output
        try:
            output.flush()
        except Exception:
            pass
        try:
            output_fd = output.fileno()
        except Exception:
            output_fd = None
        try:
            output_fn = Filename(output.name)
        except Exception:
            pass
    elif isinstance(output, int):
        output_fh = None
        output_fn = None
        output_fd = output
    elif isinstance(output, (str, Filename)):
        output_fh = None
        output_fn = Filename(output)
        output_fd = None
    else:
        raise TypeError(
            "remote_print_stack_trace(): expected file/str/int; got %s"
            % (type(output).__name__,))
    temp_file = None
    remote_fn = output_fn
    if remote_fn is None and output_fd is not None:
        remote_fn = Filename("/proc/%d/fd/%d" % (os.getpid(), output_fd))
    # Figure out whether the target process will be able to open output_fn for
    # writing.  Since the target process would need to be running as the same
    # user as this process for us to be able to attach a debugger, we can
    # simply check whether we ourselves can open the file.  Typically output
    # will be fd 1 and we will have access to write to it.  However, if we're
    # sudoed, we won't be able to re-open it via the proc symlink, even though
    # we already currently have it open.  Another case is ``output`` is a
    # file-like object that isn't a real file, e.g. a StringO.  In each case
    # we we don't have a usable filename for the remote process yet.  To
    # address these situations, we create a temporary file for the remote
    # process to write to.
    if remote_fn is None or not remote_fn.iswritable:
        if not output_fh or output_fd:
            assert remote_fn is not None
            raise OSError(errno.EACCESS, "Can't write to %s" % output_fn)
        # We can still use the /proc/$pid/fd approach with an unnamed temp
        # file.  If it turns out there are situations where that doesn't work,
        # we can switch to using a NamedTemporaryFile.
        from tempfile import TemporaryFile
        temp_file = TemporaryFile()
        remote_fn = Filename(
            "/proc/%d/fd/%d" % (os.getpid(), temp_file.fileno()))
        assert remote_fn.iswritable
    # *** Do the code injection ***
    _remote_print_stack_to_file(pid, remote_fn)
    # Copy from temp file to the requested output.
    if temp_file is not None:
        data = temp_file.read()
        temp_file.close()
        if output_fh is not None:
            output_fh.write(data)
            output_fh.flush()
        elif output_fd is not None:
            with os.fdopen(output_fd, 'w') as f:
                f.write(data)
        else:
            raise AssertionError("unreacahable")


def _remote_print_stack_to_file(pid, filename):
    inject(pid, [
        "import traceback",
        "with open(%r,'w') as f: traceback.print_stack(file=f)" % str(filename)
        ], wait=True)



# Deprecated wrapper for wait_for_debugger_to_attach().
def waitpoint(frame=None, mailto=None, background=False, timeout=86400):
    if frame is None:
        frame = _get_caller_frame()
    wait_for_debugger_to_attach(frame, mailto=mailto,
                                background=background, timeout=timeout)

breakpoint                       = debugger                          # deprecated alias
debug_statement                  = debugger                          # deprecated alias
debug_exception                  = debugger                          # deprecated alias
enable_signal_handler_breakpoint = enable_signal_handler_debugger    # deprecated alias
enable_exception_handler         = enable_exception_handler_debugger # deprecated alias



# pyflyby/_importstmt.py.
# Copyright (C) 2011, 2012, 2013, 2014 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT



import ast
from   collections              import namedtuple
from   functools                import total_ordering

from   pyflyby._flags           import CompilerFlags
from   pyflyby._format          import FormatParams, pyfill
from   pyflyby._idents          import is_identifier
from   pyflyby._parse           import PythonStatement
from   pyflyby._util            import (Inf, cached_attribute, cmp,
                                        longest_common_prefix)


from   typing                   import Dict, Tuple, Optional, Union



def read_black_config() -> Dict:
    """Read the black configuration from ``pyproject.toml``"""
    from black.files import find_pyproject_toml, parse_pyproject_toml

    pyproject_path = find_pyproject_toml((".",))

    raw_config = parse_pyproject_toml(pyproject_path) if pyproject_path else {}

    config = {}
    for key in [
        "line_length",
        "skip_magic_trailing_comma",
        "skip_string_normalization",
    ]:
        if key in raw_config:
            config[key] = raw_config[key]
    if "target_version" in raw_config:
        target_version = raw_config["target_version"]
        if isinstance(target_version, str):
            config["target_version"] = target_version
        elif isinstance(target_version, list):
            # Convert TOML list to a Python set
            config["target_version"] = set(target_version)
        else:
            raise ValueError(
                f"Invalid config for black = {target_version!r} in {pyproject_path}"
            )
    return config


class ImportFormatParams(FormatParams):
    align_imports:Union[bool, set, list, tuple, str] = True
    """
    Whether and how to align 'from modulename import aliases...'.  If ``True``,
    then the 'import' keywords will be aligned within a block.  If an integer,
    then the 'import' keyword will always be at that column.  They will be
    wrapped if necessary.
    """

    from_spaces:int = 1
    """
    The number of spaces after the 'from' keyword.  (Must be at least 1.)
    """

    separate_from_imports:bool = True
    """
    Whether all 'from ... import ...' in an import block should come after
    'import ...' statements.  ``separate_from_imports = False`` works well with
    ``from_spaces = 3``.  ('from __future__ import ...' always comes first.)
    """

    align_future:bool = False
    """
    Whether 'from __future__ import ...' statements should be aligned with
    others.  If False, uses a single space after the 'from' and 'import'
    keywords.
    """


class NonImportStatementError(TypeError):
    """
    Unexpectedly got a statement that wasn't an import.
    """

ImportSplit = namedtuple("ImportSplit",
                         "module_name member_name import_as")
"""
Representation of a single import at the token level::

  from [...]<module_name> import <member_name> as <import_as>

If <module_name> is ``None``, then there is no "from" clause; instead just::
  import <member_name> as <import_as>
"""

@total_ordering
class Import:
    """
    Representation of the desire to import a single name into the current
    namespace.

      >>> Import.from_parts(".foo.bar", "bar")
      Import('from .foo import bar')

      >>> Import("from . import foo")
      Import('from . import foo')

      >>> Import("from . import foo").fullname
      '.foo'

      >>> Import("import   foo . bar")
      Import('import foo.bar')

      >>> Import("import   foo . bar  as  baz")
      Import('from foo import bar as baz')

      >>> Import("import   foo . bar  as  bar")
      Import('from foo import bar')

      >>> Import("foo.bar")
      Import('from foo import bar')

    """

    fullname:str
    import_as:str

    def __new__(cls, arg):
        if isinstance(arg, cls):
            return arg
        if isinstance(arg, ImportSplit):
            return cls.from_split(arg)
        if isinstance(arg, (ImportStatement, PythonStatement)):
            return cls._from_statement(arg)
        if isinstance(arg, str):
            return cls._from_identifier_or_statement(arg)
        raise TypeError

    @classmethod
    def from_parts(cls, fullname, import_as):
        assert isinstance(fullname, str)
        assert isinstance(import_as, str)
        self = object.__new__(cls)
        self.fullname = fullname
        self.import_as = import_as
        return self

    @classmethod
    def _from_statement(cls, statement):
        """
        :type statement:
          `ImportStatement` or convertible (`PythonStatement`, ``str``)
        :rtype:
          `Import`
        """
        statement = ImportStatement(statement)
        imports = statement.imports
        if len(imports) != 1:
            raise ValueError(
                "Got %d imports instead of 1 in %r" % (len(imports), statement))
        return imports[0]

    @classmethod
    def _from_identifier_or_statement(cls, arg):
        """
        Parse either a raw identifier or a statement.

          >>> Import._from_identifier_or_statement('foo.bar.baz')
          Import('from foo.bar import baz')

          >>> Import._from_identifier_or_statement('import foo.bar.baz')
          Import('import foo.bar.baz')

        :rtype:
          `Import`
        """
        if is_identifier(arg, dotted=True):
            return cls.from_parts(arg, arg.split('.')[-1])
        else:
            return cls._from_statement(arg)

    @cached_attribute
    def split(self):
        """
        Split this `Import` into a ``ImportSplit`` which represents the
        token-level ``module_name``, ``member_name``, ``import_as``.

        Note that at the token level, ``import_as`` can be ``None`` to represent
        that the import statement doesn't have an "as ..." clause, whereas the
        ``import_as`` attribute on an ``Import`` object is never ``None``.

          >>> Import.from_parts(".foo.bar", "bar").split
          ImportSplit(module_name='.foo', member_name='bar', import_as=None)

          >>> Import("from . import foo").split
          ImportSplit(module_name='.', member_name='foo', import_as=None)

          >>> Import.from_parts(".foo", "foo").split
          ImportSplit(module_name='.', member_name='foo', import_as=None)

          >>> Import.from_parts("foo.bar", "foo.bar").split
          ImportSplit(module_name=None, member_name='foo.bar', import_as=None)

        :rtype:
          `ImportSplit`
        """
        if self.import_as == self.fullname:
            return ImportSplit(None, self.fullname, None)
        level = 0
        qname = self.fullname
        for level, char in enumerate(qname):
            if char != '.':
                break
        prefix = qname[:level]
        qname = qname[level:]
        if '.' in qname:
            module_name, member_name = qname.rsplit(".", 1)
        else:
            module_name = ''
            member_name = qname
        module_name = prefix + module_name
        import_as = self.import_as
        if import_as == member_name:
            import_as = None
        return ImportSplit(module_name or None, member_name, import_as)

    @classmethod
    def from_split(cls, impsplit):
        """
        Construct an `Import` instance from ``module_name``, ``member_name``,
        ``import_as``.

        :rtype:
          `Import`
        """
        impsplit = ImportSplit(*impsplit)
        module_name, member_name, import_as = impsplit
        if import_as is None:
            import_as = member_name
        if module_name is None:
            result = cls.from_parts(member_name, import_as)
        else:
            fullname = "%s%s%s" % (
                module_name,
                "" if module_name.endswith(".") else ".",
                member_name)
            result = cls.from_parts(fullname, import_as)
        # result.split will usually be the same as impsplit, but could be
        # different if the input was 'import foo.bar as baz', which we
        # canonicalize to 'from foo import bar as baz'.
        return result

    def prefix_match(self, imp):
        """
        Return the longest common prefix between ``self`` and ``imp``.

          >>> Import("import ab.cd.ef").prefix_match(Import("import ab.cd.xy"))
          ('ab', 'cd')

        :type imp:
          `Import`
        :rtype:
          ``tuple`` of ``str``
        """
        imp = Import(imp)
        n1 = self.fullname.split('.')
        n2 = imp.fullname.split('.')
        return tuple(longest_common_prefix(n1, n2))

    def replace(self, prefix, replacement):
        """
        Return a new ``Import`` that replaces ``prefix`` with ``replacement``.

          >>> Import("from aa.bb import cc").replace("aa.bb", "xx.yy")
          Import('from xx.yy import cc')

          >>> Import("from aa import bb").replace("aa.bb", "xx.yy")
          Import('from xx import yy as bb')

        :rtype:
          ``Import``
        """
        prefix_parts = prefix.split('.')
        replacement_parts = replacement.split('.')
        fullname_parts = self.fullname.split('.')
        if fullname_parts[:len(prefix_parts)] != prefix_parts:
            # No prefix match.
            return self
        fullname_parts[:len(prefix_parts)] = replacement_parts
        import_as_parts = self.import_as.split('.')
        if import_as_parts[:len(prefix_parts)] == prefix_parts:
            import_as_parts[:len(prefix_parts)] = replacement_parts
        return self.from_parts('.'.join(fullname_parts),
                               '.'.join(import_as_parts))

    @cached_attribute
    def flags(self):
        """
        If this is a __future__ import, then the compiler_flag associated with
        it.  Otherwise, 0.
        """
        if self.split.module_name == "__future__":
            return CompilerFlags(self.split.member_name)
        else:
            return CompilerFlags.from_int(0)

    @property
    def _data(self):
        return (self.fullname, self.import_as)

    def pretty_print(self, params=FormatParams()):
        return ImportStatement([self]).pretty_print(params)

    def __str__(self):
        return self.pretty_print(FormatParams(max_line_length=Inf)).rstrip()

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, str(self))

    def __hash__(self):
        return hash(self._data)

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, Import):
            return NotImplemented
        return cmp(self._data, other._data)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Import):
            return NotImplemented
        return self._data == other._data

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, other):
        if self is other:
            return False
        if not isinstance(other, Import):
            return NotImplemented
        return self._data < other._data

def _validate_alias(arg) -> Tuple[str, Optional[str]]:
    """
    Ensure each alias is a tuple (str, None|str), and return it.

    """
    assert isinstance(arg, tuple)
    # Pyright does not seem to be able to infer the length from a
    # the unpacking.
    assert len(arg) == 2
    a0, a1 = arg
    assert isinstance(a0, str)
    assert isinstance(a1, (str, type(None)))
    return arg

@total_ordering
class ImportStatement:
    """
    Token-level representation of an import statement containing multiple
    imports from a single module.  Corresponds to an ``ast.ImportFrom`` or
    ``ast.Import``.
    """

    aliases : Tuple[Tuple[str, Optional[str]],...]
    fromname : Optional[str]

    def __new__(cls, arg):
        if isinstance(arg, cls):
            return arg
        if isinstance(arg, str):
            return cls._from_str(arg)
        if isinstance(arg, PythonStatement):
            return cls._from_statement(arg)
        if isinstance(arg, (ast.ImportFrom, ast.Import)):
            return cls._from_ast_node(arg)
        if isinstance(arg, Import):
            return cls._from_imports([arg])
        if isinstance(arg, (tuple, list)) and len(arg):
            if isinstance(arg[0], Import):
                return cls._from_imports(arg)
        raise TypeError

    @classmethod
    def from_parts(cls, fromname:Optional[str], aliases:Tuple[Tuple[str, Optional[str]],...]):
        assert isinstance(aliases, tuple)
        assert len(aliases) > 0

        self = object.__new__(cls)
        self.fromname = fromname
        self.aliases = tuple(_validate_alias(a) for a in aliases)
        return self

    @classmethod
    def _from_str(cls, code:str, /):
        """
          >>> ImportStatement._from_str("from foo  import bar, bar2, bar")
          ImportStatement('from foo import bar, bar2, bar')

          >>> ImportStatement._from_str("from foo  import bar as bar")
          ImportStatement('from foo import bar as bar')

          >>> ImportStatement._from_str("from foo.bar  import baz")
          ImportStatement('from foo.bar import baz')

          >>> ImportStatement._from_str("import  foo.bar")
          ImportStatement('import foo.bar')

          >>> ImportStatement._from_str("from .foo  import bar")
          ImportStatement('from .foo import bar')

          >>> ImportStatement._from_str("from .  import bar, bar2")
          ImportStatement('from . import bar, bar2')

        :type statement:
          `PythonStatement`
        :rtype:
          `ImportStatement`
        """

        statement = PythonStatement(code)
        return cls._from_ast_node(statement.ast_node)

    @classmethod
    def _from_statement(cls, statement):
        statement = PythonStatement.from_statement(statement)
        return cls._from_ast_node(statement.ast_node)

    @classmethod
    def _from_ast_node(cls, node):
        """
        Construct an `ImportStatement` from an `ast` node.

        :rtype:
          `ImportStatement`
        """
        if isinstance(node, ast.ImportFrom):
            if node.module is None:
                module = ''
            else:
                assert isinstance(node.module, str)
                module = node.module
            fromname = '.' * node.level + module
        elif isinstance(node, ast.Import):
            fromname = None
        else:
            raise NonImportStatementError(
                    'Expected ImportStatement, got {node}'.format(node=node)
                    )
        aliases = tuple( (alias.name, alias.asname) for alias in node.names )
        return cls.from_parts(fromname, aliases)

    @classmethod
    def _from_imports(cls, imports):
        """
        Construct an `ImportStatement` from a sequence of ``Import`` s.  They
        must all have the same ``fromname``.

        :type imports:
          Sequence of `Import` s
        :rtype:
          `ImportStatement`
        """
        if not all(isinstance(imp, Import) for imp in imports):
            raise TypeError
        if not len(imports) > 0:
            raise ValueError
        module_names = set(imp.split.module_name for imp in imports)
        if len(module_names) > 1:
            raise ValueError(
                "Inconsistent module names %r" % (sorted(module_names),))
        fromname = list(module_names)[0]
        aliases = tuple(imp.split[1:] for imp in imports)
        return cls.from_parts(fromname, aliases)

    @cached_attribute
    def imports(self):
        """
        Return a sequence of `Import` s.

        :rtype:
          ``tuple`` of `Import` s
        """
        return tuple(
            Import.from_split((self.fromname, alias[0], alias[1]))
            for alias in self.aliases)

    @property
    def module(self) -> Tuple[str, ...]:
        """

        return the import module as a list of string (which would be joined by
        dot in the original import form.

        This is useful for sorting purposes

        Note that this may contain some empty string in particular with relative
        imports
        """
        if self.fromname:
            return tuple(self.fromname.split('.'))


        assert len(self.aliases) == 1, self.aliases

        return tuple(self.aliases[0][0].split('.'))

        
    def _cmp(self):
        """
        Comparison function for sorting.

        We want to sort:
            - by the root module
            - whether it is an "import ... as", or "from ... import as" import
            - then lexicographically

        """
        return (self.module[0], 0 if self.fromname is not None else 1, self.fromname)

    @cached_attribute
    def flags(self):
        """
        If this is a __future__ import, then the bitwise-ORed of the
        compiler_flag values associated with the features.  Otherwise, 0.
        """
        return CompilerFlags(*[imp.flags for imp in self.imports])

    def pretty_print(self, params=FormatParams(),
                     import_column=None, from_spaces=1):
        """
        Pretty-print into a single string.

        :type params:
          `FormatParams`
        :param modulename_ljust:
          Number of characters to left-justify the 'from' name.
        :rtype:
          ``str``
        """
        s0 = ''
        s = ''
        assert from_spaces >= 1
        if self.fromname is not None:
            s += "from%s%s " % (' ' * from_spaces, self.fromname)
            if import_column is not None:
                if len(s) > import_column:
                    # The caller wants the 'import' statement lined up left of
                    # where the current end of the line is.  So wrap it
                    # specially like this::
                    #     from foo     import ...
                    #     from foo.bar.baz \
                    #                  import ...
                    s0 = s + '\\\n'
                    s = ' ' * import_column
                else:
                    s = s.ljust(import_column)
        s += "import "
        tokens = []
        for importname, asname in self.aliases:
            if asname is not None:
                t = "%s as %s" % (importname, asname)
            else:
                t = "%s" % (importname,)
            tokens.append(t)
        res = s0 + pyfill(s, tokens, params=params)
        if params.use_black:
            return self.run_black(res, params)
        return res

    @staticmethod
    def run_black(src_contents: str, params:FormatParams) -> str:
        """Run the black formatter for the Python source code given as a string

        This is adapted from https://github.com/akaihola/darker

        """
        from black import format_str, FileMode
        from black.mode import TargetVersion

        black_config = read_black_config()
        mode = dict()
        if "line_length" in black_config:
            mode["line_length"] = (
                params.max_line_length
                if params.max_line_length
                else black_config["line_length"]
            )
        if "target_version" in black_config:
            if isinstance(black_config["target_version"], set):
                target_versions_in = black_config["target_version"]
            else:
                target_versions_in = {black_config["target_version"]}
            all_target_versions = {
                tgt_v.name.lower(): tgt_v for tgt_v in TargetVersion
            }
            bad_target_versions = target_versions_in - set(all_target_versions)
            if bad_target_versions:
                raise ValueError(
                    f"Invalid target version(s) {bad_target_versions}"
                )
            mode["target_versions"] = {
                all_target_versions[n] for n in target_versions_in
            }
        if "skip_magic_trailing_comma" in black_config:
            mode["magic_trailing_comma"] = not black_config[
                "skip_magic_trailing_comma"
            ]
        if "skip_string_normalization" in black_config:
            # The ``black`` command line argument is
            # ``--skip-string-normalization``, but the parameter for
            # ``black.Mode`` needs to be the opposite boolean of
            # ``skip-string-normalization``, hence the inverse boolean
            mode["string_normalization"] = not black_config["skip_string_normalization"]

        # The custom handling of empty and all-whitespace files below will be unnecessary if
        # https://github.com/psf/black/pull/2484 lands in Black.
        contents_for_black = src_contents
        return format_str(contents_for_black, mode=FileMode(**mode))

    @property
    def _data(self):
        return (self.fromname, self.aliases)

    def __str__(self):
        return self.pretty_print(FormatParams(max_line_length=Inf)).rstrip()

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, str(self))

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, ImportStatement):
            return NotImplemented
        return cmp(self._data, other._data)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, ImportStatement):
            return NotImplemented
        return self._data == other._data

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, other):
        if not isinstance(other, ImportStatement):
            return NotImplemented
        return self._data < other._data

    def __hash__(self):
        return hash(self._data)



from __future__ import annotations

import os
import pytest
import random
import pickle
from   shutil                   import rmtree
import subprocess
import sys
from   tempfile                 import mkdtemp
from   textwrap                 import dedent
from   contextlib               import contextmanager

from   pyflyby                  import Filename, saveframe

VERSION_INFO = sys.version_info
PYFLYBY_HOME = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
BIN_DIR = os.path.join(PYFLYBY_HOME, "bin")

exception_info_keys = {
    'exception_full_string', 'exception_object', 'exception_string',
    'exception_class_name', 'exception_class_qualname', 'traceback'}


@pytest.fixture
def tmpdir(request):
    """
    A temporary directory which is temporarily added to sys.path.
    """
    d = mkdtemp(prefix="pyflyby_test_saveframe_", suffix=".tmp")
    d = Filename(d).real
    def cleanup():
        # Unload temp modules.
        for name, module in sorted(sys.modules.items()):
            if (getattr(module, "__file__", None) or "").startswith(str(d)):
                del sys.modules[name]
        # Clean up sys.path.
        sys.path.remove(str(d))
        # Clean up directory on disk.
        rmtree(str(d))
    request.addfinalizer(cleanup)
    sys.path.append(str(d))
    return d


def load_pkl(filename):
    with open(filename, mode='rb') as f:
        data = pickle.load(f)
    return data


def writetext(filename, text, mode='w'):
    text = dedent(text)
    assert isinstance(filename, Filename)
    with open(str(filename), mode) as f:
        f.write(text)
    return filename


@contextmanager
def chdir(path):
    old_cwd = os.getcwd()
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(old_cwd)


def get_random():
    return int(random.random() * 10 ** 9)


def run_command(command):
    result = subprocess.run(command, capture_output=True)
    return result.stderr.decode('utf-8').strip().split('\n')


@contextmanager
def run_code_and_set_exception(code, exception):
    try:
        exec(code)
    except exception as err:
        if VERSION_INFO < (3, 12):
            sys.last_value = err
        else:
            sys.last_exc = err
    try:
        yield
    finally:
        delattr(sys, "last_value" if VERSION_INFO < (3, 12) else "last_exc")


def frames_metadata_checker(tmpdir, pkg_name, filename):
    """
    Check if the metadata of the frames is correctly written in the ``filename``.
    """
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4, 5}

    assert data[1]["code"] == 'raise ValueError("Error is raised")'
    assert data[1]["frame_index"] == 1
    assert data[1]["filename"] == str(
        tmpdir / pkg_name / "pkg1" / "pkg2" / "mod3.py")
    assert data[1]["lineno"] == 6
    assert data[1]["function_name"] == "func3"
    assert data[1]["function_qualname"] == "func3"
    assert data[1]["module_name"] == f"{pkg_name}.pkg1.pkg2.mod3"
    assert data[1]["frame_identifier"] == (
        f"{data[1]['filename']},{data[1]['lineno']},{data[1]['function_name']}")

    assert data[2]["code"] == 'func3()'
    assert data[2]["frame_index"] == 2
    assert data[2]["filename"] == str(tmpdir / pkg_name / "pkg1" / "mod2.py")
    assert data[2]["lineno"] == 10
    assert data[2]["function_name"] == "func2"
    assert (
        data[2]["function_qualname"] ==
        "func2" if VERSION_INFO < (3, 11) else "mod2_cls.func2")
    assert data[2]["module_name"] == f"{pkg_name}.pkg1.mod2"
    assert data[2]["frame_identifier"] == (
        f"{data[2]['filename']},{data[2]['lineno']},{data[2]['function_name']}")

    assert data[3]["code"] == 'obj.func2()'
    assert data[3]["frame_index"] == 3
    assert data[3]["filename"] == str(tmpdir / pkg_name / "mod1.py")
    assert data[3]["lineno"] == 9
    assert data[3]["function_name"] == "func2"
    assert data[3]["function_qualname"] == "func2"
    assert data[3]["module_name"] == f"{pkg_name}.mod1"
    assert data[3]["frame_identifier"] == (
        f"{data[3]['filename']},{data[3]['lineno']},{data[3]['function_name']}")

    assert data[4]["code"] == 'func2()'
    assert data[4]["frame_index"] == 4
    assert data[4]["filename"] == str(tmpdir / pkg_name / "mod1.py")
    assert data[4]["lineno"] == 14
    assert data[4]["function_name"] == "func1"
    assert data[4]["function_qualname"] == "func1"
    assert data[4]["module_name"] == f"{pkg_name}.mod1"
    assert data[4]["frame_identifier"] == (
        f"{data[4]['filename']},{data[4]['lineno']},{data[4]['function_name']}")

    assert data[5]["code"] == 'func1()'
    assert data[5]["frame_index"] == 5
    assert data[5]["filename"] == str(tmpdir / pkg_name / "__init__.py")
    assert data[5]["lineno"] == 6
    assert data[5]["function_name"] == "init_func1"
    assert data[5]["function_qualname"] == "init_func1"
    assert data[5]["module_name"] == pkg_name
    assert data[5]["frame_identifier"] == (
        f"{data[5]['filename']},{data[5]['lineno']},{data[5]['function_name']}")


def frames_metadata_checker_for_keyboard_interrupt(tmpdir, pkg_name, filename):
    """
    Check if the metadata of the frames is correctly written in the ``filename``,
    when KeyboardInterrupt exception is raised.
    """
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2}

    assert data[1]["code"] == 'os.kill(os.getpid(), signal.SIGINT)'
    assert data[1]["frame_index"] == 1
    assert data[1]["filename"] == str(
        tmpdir / pkg_name / "mod1.py")
    assert data[1]["lineno"] == 19
    assert data[1]["function_name"] == "interrupt_func"
    assert data[1]["function_qualname"] == "interrupt_func"
    assert data[1]["module_name"] == f"{pkg_name}.mod1"
    assert data[1]["frame_identifier"] == (
        f"{data[1]['filename']},{data[1]['lineno']},{data[1]['function_name']}")

    assert data[2]["code"] == 'interrupt_func()'
    assert data[2]["frame_index"] == 2
    assert data[2]["filename"] == str(tmpdir / pkg_name / "__init__.py")
    assert data[2]["lineno"] == 22
    assert data[2]["function_name"] == "init_func4"
    assert data[2]["function_qualname"] == "init_func4"
    assert data[2]["module_name"] == f"{pkg_name}"
    assert data[2]["frame_identifier"] == (
        f"{data[2]['filename']},{data[2]['lineno']},{data[2]['function_name']}")


def frames_local_variables_checker(pkg_name, filename):
    """
    Check if the local variables of the frames are correctly written in the
    ``filename``.
    """
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4, 5}

    assert set(data[1]['variables'].keys()) == {'func3_var3', 'var1', 'var2'}
    assert pickle.loads(data[1]['variables']['func3_var3']) == True
    assert pickle.loads(data[1]['variables']['var1']) == [4, 'foo', 2.4]
    assert pickle.loads(data[1]['variables']['var2']) == 'blah'

    assert set(data[2]['variables'].keys()) == {'self', 'var1', 'var2'}
    self_val = pickle.loads(data[2]['variables']['self'])
    mod2 = __import__(f"{pkg_name}.pkg1.mod2", fromlist=['dummy'], level=0)
    assert isinstance(self_val, mod2.mod2_cls)
    assert pickle.loads(data[2]['variables']['var1']) == 'foo'
    assert pickle.loads(data[2]['variables']['var2']) == (4, 9, 10)

    assert set(data[3]['variables'].keys()) == {'obj', 'var1', 'var2'}
    obj_val = pickle.loads(data[3]['variables']['obj'])
    assert isinstance(obj_val, mod2.mod2_cls)
    assert pickle.loads(data[3]['variables']['var1']) == 'func2'
    assert pickle.loads(data[3]['variables']['var2']) == 34

    assert set(data[4]['variables'].keys()) == {'func1_var2', 'var1'}
    assert pickle.loads(data[4]['variables']['func1_var2']) == 4.56
    assert pickle.loads(data[4]['variables']['var1']) == [4, 5, 2]

    assert set(data[5]['variables'].keys()) == {'var1', 'var2'}
    assert pickle.loads(data[5]['variables']['var1']) == 3
    assert pickle.loads(data[5]['variables']['var2']) == 'blah'


def frames_local_variables_checker_for_keyboard_interrupt(filename):
    """
    Check if the local variables of the frames are correctly written in the
    ``filename``, when KeyboardInterrupt exception is raised.
    """
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2}

    assert set(data[1]['variables'].keys()) == {'interrupt_var1', 'interrupt_var2'}
    assert pickle.loads(data[1]['variables']['interrupt_var1']) == 'foo bar'

    assert set(data[2]['variables'].keys()) == {'var1', 'var2'}
    assert pickle.loads(data[2]['variables']['var1']) == 'init_func4'
    assert pickle.loads(data[2]['variables']['var2']) == [3, 4]


def exception_info_checker(filename):
    """
    Check if the exception info is correctly written in the ``filename``.
    """
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys
    assert data['exception_class_name'] == 'ValueError'
    assert data['exception_class_qualname'] == 'ValueError'
    assert data['exception_full_string'] == 'ValueError: Error is raised'
    assert isinstance(data['exception_object'], ValueError)
    # Traceback shouldn't be pickled for security reasons.
    assert data['exception_object'].__traceback__ == None
    assert data['exception_string'] == 'Error is raised'


def exception_info_checker_for_keyboard_interrupt(filename):
    """
    Check if the exception info is correctly written in the ``filename``,
    when KeyboardInterrupt exception is raised.
    """
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys
    assert data['exception_class_name'] == 'KeyboardInterrupt'
    assert data['exception_class_qualname'] == 'KeyboardInterrupt'
    assert data['exception_full_string'] == 'KeyboardInterrupt: '
    assert isinstance(data['exception_object'], KeyboardInterrupt)
    # Traceback shouldn't be pickled for security reasons.
    assert data['exception_object'].__traceback__ == None
    assert data['exception_string'] == ''


def create_pkg(tmpdir):
    """
    Create a pacakge with multiple nested sub-packages and modules in ``tmpdir``.
    """
    pkg_name = f"saveframe_{int(random.random() * 10**9)}"
    os.mkdir(str(tmpdir / pkg_name))
    writetext(tmpdir / pkg_name / "__init__.py", f"""
        from {pkg_name}.mod1 import func1, interrupt_func
        def init_func1():
            var1 = 3
            var2 = 'blah'
            func1()

        def init_func2():
            pass

        def init_func3():
            var1 = 'init_func3'
            var2 = 24
            try:
                func1()
            except ValueError as err:
                raise TypeError("Chained exception") from err

        def init_func4():
            var1 = 'init_func4'
            var2 = [3, 4]
            interrupt_func()
    """)
    writetext(tmpdir / pkg_name / "mod1.py", f"""
        import os
        import signal
        from {pkg_name}.pkg1.mod2 import mod2_cls
        def func2():
            var1 = "func2"
            var2 = 34
            obj = mod2_cls()
            obj.func2()

        def func1():
            var1 = [4, 5, 2]
            func1_var2 = 4.56
            func2()

        def interrupt_func():
            interrupt_var1 = 'foo bar'
            interrupt_var2 = 3.4
            os.kill(os.getpid(), signal.SIGINT)
    """)
    os.mkdir(str(tmpdir / pkg_name / "pkg1"))
    writetext(tmpdir / pkg_name / "pkg1" / "__init__.py", "")
    writetext(tmpdir / pkg_name / "pkg1" / "mod2.py", f"""
        from {pkg_name}.pkg1.pkg2.mod3 import func3
        class mod2_cls:
            def __init__(self):
                pass
            def func2(self):
                 var1 = 'foo'
                 var2 = (4, 9, 10)
                 var3 = lambda x: x+1
                 func3()
    """)
    os.mkdir(str(tmpdir/ pkg_name / "pkg1" / "pkg2"))
    writetext(tmpdir / pkg_name / "pkg1" / "pkg2" / "__init__.py", "")
    writetext(tmpdir / pkg_name / "pkg1" / "pkg2" / "mod3.py", """
        def func3():
            var1 = [4, 'foo', 2.4]
            var2 = 'blah'
            func3_var3 = True
            raise ValueError("Error is raised")
    """)
    return pkg_name


def test_saveframe_invalid_filename_1(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / pkg_name))
    err_msg = (f"{str(tmpdir / pkg_name)!a} is an already existing directory. "
               f"Please pass a different filename.")
    assert str(err.value) == err_msg


def test_saveframe_invalid_filename_2(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_dir_{get_random()}" / f"saveframe_{get_random()}.pkl")
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(FileNotFoundError) as err:
            saveframe(filename=filename)
    err_msg = (f"Error while saving the frames to the file: {filename!a}. Error: "
               f"FileNotFoundError(2, 'No such file or directory')")
    assert str(err.value) == err_msg


def test_saveframe_invalid_frames_1(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      frames="foo.py:")
    err_msg = ("Error while validating frame: 'foo.py:'. The correct syntax for "
               "a frame is 'file_regex:line_no:function_name' but frame 'foo.py:' "
               "contains 1 ':'.")
    assert str(err.value) == err_msg


def test_saveframe_invalid_frames_2(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      frames=":12:func1")
    err_msg = ("Error while validating frame: ':12:func1'. The filename / file "
               "regex must be passed in a frame.")
    assert str(err.value) == err_msg


def test_saveframe_invalid_frames_3(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      frames="file.py:12:func1,file1.py::")
    err_msg = ("Error while validating frames: 'file.py:12:func1,file1.py::'. "
               "If you want to pass multiple frames, pass a list/tuple of frames "
               "like ['file.py:12:func1', 'file1.py::'] rather than a comma "
               "separated string of frames.")
    assert str(err.value) == err_msg


def test_saveframe_invalid_frames_4(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      frames=["file.py:12:func1", "file1.py::,file2.py:34:func2"])
    err_msg = (
        "Invalid frame: 'file1.py::,file2.py:34:func2' in frames: ['file.py:12:func1', "
        "'file1.py::,file2.py:34:func2'] as it contains character ','. If you are "
        "trying to pass multiple frames, pass them as separate items in the list.")
    assert str(err.value) == err_msg


def test_saveframe_invalid_frames_5(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      frames="file.py:12:func1..file2.py::..")
    err_msg = ("Error while validating frames: 'file.py:12:func1..file2.py::..'. "
               "If you want to pass a range of frames, the correct syntax is "
               "'first_frame..last_frame'")
    assert str(err.value) == err_msg


def test_saveframe_invalid_frames_6(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      frames="file.py:foo:func1")
    err_msg = ("Error while validating frame: 'file.py:foo:func1'. The line "
               "number 'foo' can't be converted to an integer.")
    assert str(err.value) == err_msg


def test_saveframe_variables_and_exclude_variables(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(variables="foo", exclude_variables="bar")
    err_msg = "Cannot pass both `variables` and `exclude_variables` parameters."
    assert str(err.value) == err_msg


def test_saveframe_invalid_variables_1(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      variables="var1,var2")
    err_msg = ("Error while validating variables: 'var1,var2'. If you want to pass "
               "multiple variable names, pass a list/tuple of names like ['var1', "
               "'var2'] rather than a comma separated string of names.")
    assert str(err.value) == err_msg


def test_saveframe_invalid_variables_2(tmpdir, caplog):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                  variables=["var1", "1var2"])
    log_messages = [record.message for record in caplog.records]
    warning_msg = ("Invalid variable names: ['1var2']. Skipping these variables "
                   "and continuing.")
    assert warning_msg in log_messages


def test_saveframe_invalid_variables_3(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(TypeError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      variables=1)
    err_msg = ("Variables '1' must be of type list, tuple or string (for a single "
               "variable), not '<class 'int'>'")
    assert str(err.value) == err_msg


def test_saveframe_invalid_exclude_variables_1(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(ValueError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      exclude_variables="var1,var2")
    err_msg = ("Error while validating variables: 'var1,var2'. If you want to pass "
               "multiple variable names, pass a list/tuple of names like ['var1', "
               "'var2'] rather than a comma separated string of names.")
    assert str(err.value) == err_msg


def test_saveframe_invalid_exclude_variables_2(tmpdir, caplog):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                  exclude_variables=["var1", "1var2"])
    log_messages = [record.message for record in caplog.records]
    warning_msg = ("Invalid variable names: ['1var2']. Skipping these variables "
                   "and continuing.")
    assert warning_msg in log_messages


def test_saveframe_invalid_exclude_variables_3(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with pytest.raises(TypeError) as err:
            saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                      exclude_variables=1)
    err_msg = ("Variables '1' must be of type list, tuple or string (for a single "
               "variable), not '<class 'int'>'")
    assert str(err.value) == err_msg


def test_saveframe_no_error_raised(tmpdir):
    if hasattr(sys, "last_value"):
        delattr(sys, "last_value")
    if hasattr(sys, "last_exc"):
        delattr(sys, "last_exc")
    pkg_name = create_pkg(tmpdir)
    with pytest.raises(RuntimeError) as err:
        exec(f"from {pkg_name} import init_func2; init_func2()")
        saveframe()
    err_msg = ("No exception is raised currently for which to save the frames. "
               "Make sure that an uncaught exception is raised before calling "
               "the `saveframe` function.")
    assert str(err.value) == err_msg


def test_saveframe_frame_format_1(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
        # Format: 'filename:line_no:func_name'
        saveframe(filename=filename, frames="pkg1/mod2.py:10:func2")
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {2}
        assert data[2]["filename"] == str(tmpdir / pkg_name / "pkg1" / "mod2.py")
        assert data[2]["function_name"] == "func2"

        # Format: 'filename::'
        saveframe(filename=filename, frames="mod1.py::")
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {3, 4}
        assert data[3]["filename"] == str(tmpdir / pkg_name / "mod1.py")
        assert data[3]["function_name"] == "func2"

        # Format: 'filename::func_name'
        saveframe(filename=filename, frames=f"{pkg_name}/mod1.py::func1")
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {4}
        assert data[4]["filename"] == str(tmpdir / pkg_name / "mod1.py")
        assert data[4]["function_name"] == "func1"


def test_saveframe_frame_format_2(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = str(tmpdir / f"saveframe_{get_random()}.pkl")

        saveframe(filename=filename, frames=["__init__.py::", "pkg1/mod2.py:10:func2"])
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {2, 5}
        assert data[2]["filename"] == str(tmpdir / pkg_name / "pkg1" / "mod2.py")
        assert (
            data[2]["function_qualname"] ==
            "func2" if VERSION_INFO < (3, 11) else "mod2_cls.func2")

        assert data[5]["filename"] == str(tmpdir / pkg_name / "__init__.py")
        assert data[5]["function_qualname"] == "init_func1"

        saveframe(filename=filename, frames=["pkg1/pkg2/mod3.py:6:", "mod1::"])
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {1, 3, 4}
        assert data[1]["filename"] == str(tmpdir / pkg_name / "pkg1" / "pkg2" / "mod3.py")
        assert data[1]["function_qualname"] == "func3"


def test_saveframe_frame_format_3(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
        saveframe(filename=filename, frames=3)
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {1, 2, 3}

        saveframe(filename=filename, frames=5)
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4, 5}


def test_saveframe_frame_format_4(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
        saveframe(filename=filename, frames="pkg1/mod2.py::..__init__.py:6:init_func1")
        data = load_pkl(filename)
        assert set(data.keys()) == exception_info_keys | {2, 3, 4, 5}

        with pytest.raises(ValueError) as err:
            saveframe(filename=filename,
                      frames="pkg1/mod3.py::..__init__.py:6:init_func1")
        err_msg = "No frame in the traceback matched the frame: 'pkg1/mod3.py::'"
        assert str(err.value) == err_msg


def test_saveframe_frame_format_5(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                             frames=f"{str(tmpdir)}/{pkg_name}/mod1.py::..")
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4}


def test_saveframe_variables(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                             frames=5, variables=['var1', 'var2'])
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4, 5}

    assert set(data[1]["variables"].keys()) == {"var1", "var2"}
    assert set(data[2]["variables"].keys()) == {"var1", "var2"}
    assert set(data[3]["variables"].keys()) == {"var1", "var2"}
    assert set(data[4]["variables"].keys()) == {"var1"}
    assert set(data[5]["variables"].keys()) == {"var1", "var2"}


def test_saveframe_exclude_variables(tmpdir, caplog):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = saveframe(filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
                             frames=5, exclude_variables=['var1', 'var2'])
    log_messages = [record.message for record in caplog.records]
    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4, 5}

    assert set(data[1]["variables"].keys()) == {"func3_var3"}
    assert set(data[2]["variables"].keys()) == {"self"}
    assert set(data[3]["variables"].keys()) == {"obj"}
    assert set(data[4]["variables"].keys()) == {"func1_var2"}
    assert set(data[5]["variables"].keys()) == set()
    qualname = "func2" if VERSION_INFO < (3, 11) else "mod2_cls.func2"
    warning_msg = (
        f"Cannot pickle variable: 'var3' for frame: 'File: {str(tmpdir)}/{pkg_name}"
        f"/pkg1/mod2.py, Line: 10, Function: {qualname}'.")
    assert warning_msg in "\n".join(log_messages)


def test_saveframe_defaults(tmpdir, caplog):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        with chdir(tmpdir):
            filename  = saveframe()
            log_messages = [record.message for record in caplog.records]
    # Test that saveframe.pkl file in the current working directory is used by
    # default.
    info_message = (f"Filename is not passed explicitly using the `filename` "
                    f"parameter. The frame info will be saved in the file: "
                    f"'{str(tmpdir)}/saveframe.pkl'.")
    assert info_message in log_messages
    info_message = ("`frames` parameter is not passed explicitly. The first frame "
                    "from the bottom will be saved by default.")
    assert info_message in log_messages
    assert os.path.basename(filename) == 'saveframe.pkl'
    data = load_pkl(filename)
    # Test that only first frame from the bottom (index = 1) is stored in the
    # data by default.
    assert set(data.keys()) == exception_info_keys | {1}


def test_saveframe_frame_metadata(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = saveframe(
            filename=str(tmpdir / f"saveframe_{get_random()}.pkl"), frames=5)
    frames_metadata_checker(tmpdir, pkg_name, filename)


def test_saveframe_local_variables_data(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = saveframe(
            filename=str(tmpdir / f"saveframe_{get_random()}.pkl"), frames=5)
    frames_local_variables_checker(pkg_name, filename)


def test_saveframe_exception_info(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func1; init_func1()"
    with run_code_and_set_exception(code, ValueError):
        filename = saveframe(
            filename=str(tmpdir / f"saveframe_{get_random()}.pkl"), frames=0)
    exception_info_checker(filename)


def test_saveframe_chained_exceptions(tmpdir):
    pkg_name = create_pkg(tmpdir)
    code = f"from {pkg_name} import init_func3; init_func3()"
    with run_code_and_set_exception(code, TypeError):
        filename = saveframe(
            filename=str(tmpdir / f"saveframe_{get_random()}.pkl"),
            frames=['__init__.py::init_func3', f'.*/{pkg_name}/.*::'])
    data = load_pkl(filename)

    assert 1 in set(data.keys())
    assert data[1]["code"] == 'raise TypeError("Chained exception") from err'
    assert data[1]["frame_index"] == 1
    assert data[1]["filename"] == str(
        tmpdir / pkg_name / "__init__.py")
    assert data[1]["lineno"] == 17
    assert data[1]["function_name"] == "init_func3"
    assert data[1]["function_qualname"] == "init_func3"
    assert data[1]["module_name"] == f"{pkg_name}"
    assert data[1]["frame_identifier"] == (
        f"{data[1]['filename']},{data[1]['lineno']},{data[1]['function_name']}")
    assert len(set(data.keys()) - exception_info_keys) == 5


def test_keyboard_interrupt_frame_metadata(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    code = dedent(f"""
    import sys
    from pyflyby import saveframe
    sys.path.append('{tmpdir}')
    from {pkg_name} import init_func4
    try:
        init_func4()
    except KeyboardInterrupt as err:
        if {VERSION_INFO[:2]} < (3, 12):
            sys.last_value = err
        else:
            sys.last_exc = err
    saveframe(filename='{filename}', frames=2)
    """)
    run_command(["python", "-c", code])
    frames_metadata_checker_for_keyboard_interrupt(tmpdir, pkg_name, filename)


def test_keyboard_interrupt_local_variables_data(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    code = dedent(f"""
        import sys
        from pyflyby import saveframe
        sys.path.append('{tmpdir}')
        from {pkg_name} import init_func4
        try:
            init_func4()
        except KeyboardInterrupt as err:
            if {VERSION_INFO[:2]} < (3, 12):
                sys.last_value = err
            else:
                sys.last_exc = err
        saveframe(filename='{filename}', frames=2)
        """)
    run_command(["python", "-c", code])
    frames_local_variables_checker_for_keyboard_interrupt(filename)


def test_keyboard_interrupt_exception_info(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    code = dedent(f"""
        import sys
        from pyflyby import saveframe
        sys.path.append('{tmpdir}')
        from {pkg_name} import init_func4
        try:
            init_func4()
        except KeyboardInterrupt as err:
            if {VERSION_INFO[:2]} < (3, 12):
                sys.last_value = err
            else:
                sys.last_exc = err
        saveframe(filename='{filename}', frames=0)
        """)
    run_command(["python", "-c", code])
    exception_info_checker_for_keyboard_interrupt(filename)


def test_saveframe_cmdline_no_exception():
    command = [BIN_DIR+"/saveframe", "python", "-c", "import os;"]
    err = run_command(command)
    err_msg = "Error: No exception is raised by the program: 'python -c import os;'"
    assert err_msg in err


def test_saveframe_cmdline_variables_and_exclude_variables():
    command = [BIN_DIR + "/saveframe", "--variables", "foo,bar",
               "--exclude_variables", "var", "python", "-c", "import os;"]
    err = run_command(command)
    err_msg = ("ValueError: Cannot pass both --variables and --exclude_variables "
               "arguments.")
    assert err_msg in err


def test_saveframe_cmdline_invalid_command_1():
    command = [BIN_DIR+"/saveframe", "python", "-c"]
    err = run_command(command)
    err_msg = "Error: Please pass a valid script / command to run!"
    assert err_msg in err


def test_saveframe_cmdline_invalid_command_2():
    command = [BIN_DIR+"/saveframe", "python"]
    err = run_command(command)
    err_msg = "Error: Please pass a valid script / command to run!"
    assert err_msg in err


def test_saveframe_cmdline_invalid_command_3():
    command = [BIN_DIR+"/saveframe"]
    err = run_command(command)
    err_msg = "Error: Please pass a valid script / command to run!"
    assert err_msg in err


def test_saveframe_cmdline_variables(tmpdir):
    pkg_name = create_pkg(tmpdir)
    tmp_mod = writetext(tmpdir / f"tmp_mod_{get_random()}.py", f"""
            import sys
            sys.path.append('{tmpdir}')
            from {pkg_name} import init_func1
            init_func1()
        """)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR + "/saveframe", "--filename", filename, "--frames", "5",
        "--variables", "var1, var2", "python", str(tmp_mod)]
    run_command(command)

    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4, 5}

    assert set(data[1]["variables"].keys()) == {"var1", "var2"}
    assert set(data[2]["variables"].keys()) == {"var1", "var2"}
    assert set(data[3]["variables"].keys()) == {"var1", "var2"}
    assert set(data[4]["variables"].keys()) == {"var1"}
    assert set(data[5]["variables"].keys()) == {"var1", "var2"}


def test_saveframe_cmdline_exclude_variables(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR + "/saveframe", "--filename", filename, "--frames", "5",
        "--exclude_variables", "var1,var2", "python", "-c",
        f"import sys; sys.path.append('{tmpdir}'); from {pkg_name} import "
        f"init_func1; init_func1()"]
    run_command(command)

    data = load_pkl(filename)
    assert set(data.keys()) == exception_info_keys | {1, 2, 3, 4, 5}

    assert set(data[1]["variables"].keys()) == {"func3_var3"}
    assert set(data[2]["variables"].keys()) == {"self"}
    assert set(data[3]["variables"].keys()) == {"obj"}
    assert set(data[4]["variables"].keys()) == {"func1_var2"}
    assert set(data[5]["variables"].keys()) == set()


def test_saveframe_cmdline_frame_metadata(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR+"/saveframe", "--filename", filename, "--frames", "5",
        "python", "-c",
        f"import sys; sys.path.append('{tmpdir}'); from {pkg_name} import "
        f"init_func1; init_func1()"]
    run_command(command)
    frames_metadata_checker(tmpdir, pkg_name, filename)


def test_saveframe_cmdline_local_variables_data(tmpdir):
    pkg_name = create_pkg(tmpdir)
    tmp_mod = writetext(tmpdir / f"tmp_mod_{get_random()}.py", f"""
        import sys
        sys.path.append('{tmpdir}')
        from {pkg_name} import init_func1
        init_func1()
    """)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR + "/saveframe", "--filename", filename, "--frames", "5",
        "python", str(tmp_mod)]
    run_command(command)
    frames_local_variables_checker(pkg_name, filename)


def test_saveframe_cmdline_exception_info(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR + "/saveframe", "--filename", filename, "--frames", "0",
        "python", "-c",
        f"import sys; sys.path.append('{tmpdir}'); from {pkg_name} import "
        f"init_func1; init_func1()"]
    run_command(command)
    exception_info_checker(filename)


def test_saveframe_cmdline_keyboard_interrupt_frame_metadata(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR+"/saveframe", "--filename", filename, "--frames", "2",
        "python", "-c",
        f"import sys; sys.path.append('{tmpdir}'); from {pkg_name} import "
        f"init_func4; init_func4()"]
    run_command(command)
    frames_metadata_checker_for_keyboard_interrupt(tmpdir, pkg_name, filename)


def test_saveframe_cmdline_keyboard_interrupt_local_variables_data(tmpdir):
    pkg_name = create_pkg(tmpdir)
    tmp_mod = writetext(tmpdir / f"tmp_mod_{get_random()}.py", f"""
        import sys
        sys.path.append('{tmpdir}')
        from {pkg_name} import init_func4
        init_func4()
    """)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR + "/saveframe", "--filename", filename, "--frames", "2",
        "python", str(tmp_mod)]
    run_command(command)
    frames_local_variables_checker_for_keyboard_interrupt(filename)


def test_saveframe_cmdline_keyboard_interrupt_exception_info(tmpdir):
    pkg_name = create_pkg(tmpdir)
    filename = str(tmpdir / f"saveframe_{get_random()}.pkl")
    command = [
        BIN_DIR + "/saveframe", "--filename", filename, "--frames", "0",
        "python", "-c",
        f"import sys; sys.path.append('{tmpdir}'); from {pkg_name} import "
        f"init_func4; init_func4()"]
    run_command(command)
    exception_info_checker_for_keyboard_interrupt(filename)
